data(dfs)
data(const.Q)
data(var.n)

STM.simulation <- function(time.series, n1, n2,
                           c.bound=c(2,3,4),
                           right.width=30,
                           min.left.width=right.width,
                           min.width=right.width/3,
                           max.width=200,
                           sign.level=0.001,
                           autocorrelations="no"
                           ){

    time.series1 <- time.series[,1]
    time.series2 <- time.series[,2]

    length.series <- length(time.series1)

    r <- right.width
    ell.min <- min.left.width
    n.min <- min.width
    n.max <- max.width

    if(autocorrelations=="high.negative"){
        c.q.sim   <- const.Q[,1]
        c.q.model <- function(n){
            c.q.sim[150]
        }
    }
    if(autocorrelations=="moderate.negative"){
        c.q.sim   <- const.Q[,2]
        c.q.model <- function(n){
             c.q.sim[150]
        }
    }
    if(autocorrelations=="small.negative"){
        c.q.sim   <- const.Q[,3]
        c.q.model <- function(n){
             c.q.sim[150]
        }
    }
    if(autocorrelations=="no"){
        c.q.sim   <- const.Q[,4]
        c.q.model <- function(n){
            1.21 * (n/(n+0.44))
        }
    }
    if(autocorrelations=="small.positive"){
        c.q.sim   <- const.Q[,5]
        c.q.model <- function(n){
             c.q.sim[150]
        }
    }
    if(autocorrelations=="moderate.positive"){
        c.q.sim   <- const.Q[,6]
        c.q.model <- function(n){
             c.q.sim[150]
        }
    }
    if(autocorrelations=="high.positive"){
        c.q.sim   <- const.Q[,7]
        c.q.model <- function(n){
             c.q.sim[150]
        }
    }

    if (autocorrelations == "high.negative") {
        emp.var <- var.n[, 1]
    }
    if (autocorrelations == "moderate.negative") {
        emp.var <- var.n[, 2]
    }
    if (autocorrelations == "small.negative") {
        emp.var <- var.n[, 3]
    }
    if (autocorrelations == "no") {
        emp.var <- var.n[, 4]
    }
    if (autocorrelations == "small.positive") {
        emp.var <- var.n[, 5]
    }
    if (autocorrelations == "moderate.positive") {
        emp.var <- var.n[, 6]
    }
    if (autocorrelations == "high.positive") {
        emp.var <- var.n[, 7]
    }

    get.Q.adj <- function(y){
        n <- length(y)
        if(n <= 150){
            c.q <- c.q.sim[n]
        } else {
            c.q <- c.q.model(n)
        }
        h <- numeric(n-2)
        for(j in 1:(n-2)){
            h[j] <- abs(y[j+1] - ((y[j]+y[j+2])/2))
        }
        h <- sort(h)
        out <- c.q * sort(h)[max(1,floor(0.5*(n-2)))]
        return(out)
    }

    get.Ogk <- function(x, noise.sd.est.s){
        D.Ogk <- diag(noise.sd.est.s)
        Y.Ogk <- x%*%solve(D.Ogk)
        R.Ogk <- diag(1, nrow=ncol(x))
        Rij.Ogk <- function(x){
            ((get.Q.adj(apply(x,1,sum)))^2 -  (get.Q.adj(apply(x,1,diff)))^2)/4
        }
        comparisons <- t(combn(1:ncol(x),2))
        for(i in 1:nrow(comparisons)){
            column1 <- comparisons[i,1]
            column2 <- comparisons[i,2]
            R.Ogk[column1,column2] <-  R.Ogk[column2,column1] <- Rij.Ogk(cbind(Y.Ogk[,column1], Y.Ogk[,column2]))
        }
        Eigen.R.Ogk <- eigen(R.Ogk)
        E.Ogk <-  Eigen.R.Ogk$vectors
        Lambda.Ogk <- Eigen.R.Ogk$values
        A.Ogk <- D.Ogk%*% E.Ogk
        pre.Gamma.Ogk <- x%*%solve(t(A.Ogk))
        Gamma.Ogk <- diag(apply(pre.Gamma.Ogk, 2, get.Q.adj)^2)
        Ogk.est <- A.Ogk %*% Gamma.Ogk %*% t(A.Ogk)
        return(Ogk.est)
    }

    corr.RM.slope.model <- function(ratio, rho){
        out <- (rho*0.08986669) + ((rho^3)*0.01312361) - (rho*ratio*1.17617121) + (rho*(ratio^2)*3.57618663) -
        (rho*(ratio^3)*1.72785517) - ((rho^3)*ratio*0.18067041)+ ((rho^3)*(ratio^2)*0.58289707) -
        ((rho^3)*(ratio^3)*0.19580252)
        return(out)
    }

    get.critval <- function(ell, r, sign.level){
        if(ell <= 100) {
            ell <- ell - ell%%5
            r <- r - r%%5
            degree.of.freedom <- dfs[ell/5, r/5]
            critval <- qt(p = 1 - (sign.level/2), df = degree.of.freedom)
        }
        else {
            critval <- qnorm(p = 1 - (sign.level/2))
        }
        return(critval)
    }

    get.B.t <- function(y) {
        n <- length(y)
        j <- 1:n
        B.t <- matrix(NA, n, n)
        for (i in j) {
            B.t[j[j < i], i] <- B.t[i, j[j < i]] <- (y[i] - y[j[j <
                i]])/(i - j[j < i])
        }
        return(B.t)
    }

    get.beta.RM <- function(B.t) {
        beta.i <- apply(B.t, 1, median, na.rm = T)
        beta.RM <- median(beta.i, na.rm = T)
        return(beta.RM)
    }

    get.B1 <- function(y, ell) {
        n <- length(y)
        r <- n - ell
        I.left <- 1:ell
        I.right <- (ell + 1):n
        B1 <- matrix(NA, ell, r)
        for (i in I.left) {
            B1[i, I.right - ell] <- (y[i] - y[I.right])/(i -
                I.right)
        }
        return(B1)
    }

    scarm.test.statistic <- function(y, B, ell, r, noise.sd.est){
        n <- ell + r
        left.sample <- y[1:ell]
        right.sample <- y[(ell + 1):n]
        B.t.left <- B[1:ell, 1:ell]
        B.t.right <- B[(ell + 1):n, (ell + 1):n]
        beta.left <- get.beta.RM(B.t.left)
        beta.right <- get.beta.RM(B.t.right)
        d.t <- beta.left - beta.right
        v.left <- emp.var[ell]
        v.right <- emp.var[r]
        var.t <- noise.sd.est^2 * (v.left + v.right)
        test.statistic <- d.t/sqrt(var.t)
        return(list(test.statistic = test.statistic, beta.left = beta.left,
            beta.right = beta.right))
    }

    max.n <- max(n1,n2)
    I.win1 <- (max.n - n1 + 1):max.n
    I.win2 <- (max.n - n2 + 1):max.n
    y1 <- time.series1[I.win1]
    y2 <- time.series2[I.win2]
    B1 <- get.B.t(y1)
    B2 <- get.B.t(y2)

    stm.statistics <- c()
    break.check <- rep(0,length(c.bound))
    time.points <- rep(NA,length(c.bound))

    for(i in max.n:length.series){

        if(n1 < ell.min + r){
            beta.est1 <- get.beta.RM(B1)
            noise.sd.est1 <- get.Q.adj(y1)
            if(n1==n.max){
                b.new1 <- (time.series1[i + 1] - time.series1[i + c((-n1 + 2):0)])/(i + 1 - (i + c((-n1 + 2):0)))
                B1 <- cbind(rbind(B1[-1, -1], b.new1), c(b.new1,NA))
                y1 <- time.series1[(i - n1 + 2):(i + 1)]
            } else {
                b.new1 <- (time.series1[i + 1] - time.series1[i + c((-n1 + 1):0)])/(i + 1 - (i + c((-n1 + 1):0)))
                B1 <- cbind(rbind(B1, b.new1), c(b.new1, NA))
                y1 <- time.series1[(i - n1 + 1):(i + 1)]
                n1 <- n1 + 1
            }
        } else {
            ell1 <- n1 - r
            noise.sd.est1 <- get.Q.adj(y1)
            critval1 <- get.critval(ell = ell1, r = r, sign.level = sign.level)
            scarm.test1 <- scarm.test.statistic(y = y1, B = B1, ell = ell1, r = r, noise.sd.est = noise.sd.est1)
            scarm.statistic1 <- scarm.test1$test.statistic
            slope.diff1 <- scarm.test1$beta.left - scarm.test1$beta.right
            if(abs(scarm.test1$test.statistic) > critval1){
                B1 <- B1[(n1 - n.min + 1):n1, (n1 - n.min + 1):n1]
                beta.est1 <- get.beta.RM(B1)
                n1 <- n.min
                b.new1 <- (time.series1[i + 1] - time.series1[i + c((-n1 + 1):0)])/(i + 1 - (i + c((-n1 + 1):0)))
                B1 <- cbind(rbind(B1, b.new1), c(b.new1, NA))
                y1 <- time.series1[(i - n1 + 1):(i + 1)]
                n1 <- n1 + 1
            } else {
                beta.est1 <- get.beta.RM(B1)
                if(n1==n.max){
                    b.new1 <- (time.series1[i + 1] - time.series1[i + c((-n1 + 2):0)])/(i + 1 - (i + c((-n1 + 2):0)))
                    B1 <- cbind(rbind(B1[-1, -1], b.new1), c(b.new1, NA))
                    y1 <- time.series1[(i - n1 + 2):(i + 1)]
                } else {
                    b.new1 <- (time.series1[i + 1] - time.series1[i + c((-n1 + 1):0)])/(i + 1 - (i + c((-n1 + 1):0)))
                    B1 <- cbind(rbind(B1, b.new1), c(b.new1, NA))
                    y1 <- time.series1[(i - n1 + 1):(i + 1)]
                    n1 <- n1 + 1
                }
            }
        }

        if(n2 < ell.min + r){
            beta.est2 <- get.beta.RM(B2)
            noise.sd.est2 <- get.Q.adj(y2)
            if(n2==n.max){
                b.new2 <- (time.series2[i + 1] - time.series2[i + c((-n2 + 2):0)])/(i + 1 - (i + c((-n2 + 2):0)))
                B2 <- cbind(rbind(B2[-1, -1], b.new2), c(b.new2,NA))
                y2 <- time.series2[(i - n2 + 2):(i + 1)]
            } else {
                b.new2 <- (time.series2[i + 1] - time.series2[i + c((-n2 + 1):0)])/(i + 1 - (i + c((-n2 + 1):0)))
                B2 <- cbind(rbind(B2, b.new2), c(b.new2, NA))
                y2 <- time.series2[(i - n2 + 1):(i + 1)]
                n2 <- n2 + 1
            }
        } else {
            ell2 <- n2 - r
            noise.sd.est2 <- get.Q.adj(y2)
            critval2 <- get.critval(ell = ell2, r = r, sign.level = sign.level)
            scarm.test2 <- scarm.test.statistic(y = y2, B = B2, ell = ell2, r = r, noise.sd.est = noise.sd.est2)
            scarm.statistic2 <- scarm.test2$test.statistic
            slope.diff2 <- scarm.test2$beta.left - scarm.test2$beta.right
            if(abs(scarm.test2$test.statistic) > critval2){
                B2 <- B2[(n2 - n.min + 1):n2, (n2 - n.min + 1):n2]
                beta.est2 <- get.beta.RM(B2)
                n2 <- n.min
                b.new2 <- (time.series2[i + 1] - time.series2[i + c((-n2 + 1):0)])/(i + 1 - (i + c((-n2 + 1):0)))
                B2 <- cbind(rbind(B2, b.new2), c(b.new2, NA))
                y2 <- time.series2[(i - n2 + 1):(i + 1)]
                n2 <- n2 + 1
            } else {
                beta.est2 <- get.beta.RM(B2)
                if(n2==n.max){
                    b.new2 <- (time.series2[i + 1] - time.series2[i + c((-n2 + 2):0)])/(i + 1 - (i + c((-n2 + 2):0)))
                    B2 <- cbind(rbind(B2[-1, -1], b.new2), c(b.new2, NA))
                    y2 <- time.series2[(i - n2 + 2):(i + 1)]
                } else {
                    b.new2 <- (time.series2[i + 1] - time.series2[i + c((-n2 + 1):0)])/(i + 1 - (i + c((-n2 + 1):0)))
                    B2 <- cbind(rbind(B2, b.new2), c(b.new2, NA))
                    y2 <- time.series2[(i - n2 + 1):(i + 1)]
                    n2 <- n2 + 1
                }
            }
        }

        var.beta1 <- noise.sd.est1^2 * emp.var[n1]
        var.beta2 <- noise.sd.est2^2 * emp.var[n2]

        ratio <- min(n1,n2)/max(n1,n2)
        rho <- cov2cor(get.Ogk(time.series[(i-max(n1,n2)+1):i,], noise.sd.est.s=c(noise.sd.est1,noise.sd.est2)))[1,2]
        corr.beta <- corr.RM.slope.model(ratio, rho)
        cov.beta <- corr.beta*sqrt(var.beta1*var.beta2)
        var.beta.diff <- var.beta1 + var.beta2 - (2*cov.beta)

        stm.stat <- (beta.est1 + beta.est2)/sqrt(var.beta.diff)
        stm.statistics[i] <- stm.stat

        for(j in 1:length(c.bound)){
            if(abs(stm.stat) > c.bound[j]){
            # STM finds incoherence!
                if(break.check[j]==0){
                    time.points[j] <- i
                    break.check[j] <- 1
                }
            }
        }
        if(all(break.check==1)){
            break
        }
    }
    return(time.points-max.n+1)
}
